module colorize(t3x, iostream, string);

object	t[t3x], str[string];

const	BUFLEN	= 1024;
const	MAXWORD	= 128;

object	In[iostream], Out[iostream];
var	Inbuf::BUFLEN, Outbuf::BUFLEN;
var	Rejected;
var	Line;
var	C_numlit,
	C_strlit,
	C_comment,
	C_def,
	C_text,
	C_bg,
	C_href;

init() do
	In.open("sol.sol", Inbuf, BUFLEN, iostream.FREAD);
	Out.open("hypersol.html", Outbuf, BUFLEN, iostream.FWRITE);
	Rejected := 0;
	Line := 1;
	C_text :=     "t";
	C_numlit :=   "n";
	C_strlit :=   "s";
	C_comment :=  "c";
	C_def :=      "d";
end

fini() do
	In.close();
	Out.close();
end

error(s) do
	t.write(T3X.SYSERR, "Colorize: error: ", 17);
	t.write(T3X.SYSERR, s, str.length(s));
	t.write(T3X.SYSERR, s, t.newline(s));
	halt 1;
end

readword(s, dlm) do var c, i;
	if (Rejected) do
		Out.wrch(Rejected);
		Rejected := 0;
	end
	c := In.rdch();
	ie (dlm) do
		while (c = dlm) do
			Out.wrch(c);
			c := In.rdch();
		end
	end
	else do
		while (c = ' ' \/ c = '\n' \/ c = '\r') do
			Out.wrch(c);
			c := In.rdch();
		end
	end
	i := 0;
	ie (dlm) do
		while (\In.eof() /\ c \= dlm) do
			if (i > MAXWORD) error("word too long");
			s::i := c;
			i := i+1;
			c := In.rdch();
		end
	end
	else do
		while (\In.eof() /\ c \= ' ' /\ c \= '\n' /\ c \= '\r') do
			if (i > MAXWORD) error("word too long");
			s::i := c;
			i := i+1;
			c := In.rdch();
		end
	end
	Rejected := c;
	s::i := 0;
end

htmlize(s) do
	var	i;

	i := 0;
	while (s::i) do
		ie (s::i = '<') Out.writes("&lt;");
		else ie (s::i = '>') Out.writes("&gt;");
		else ie (s::i = '&') Out.writes("&amp;");
		else Out.wrch(s::i);
		i := i+1;
	end
	return s;
end

linkify(s) do
	var	i;

	i := 0;
	while (s::i) do
		ie (s::i = '?') Out.wrch('Q');
		else ie (s::i = '"') Out.wrch('~');
		else ie (s::i = '<') Out.writes("&lt;");
		else ie (s::i = '>') Out.writes("&gt;");
		else ie (s::i = '&') Out.writes("&amp;");
		else Out.wrch(s::i);
		i := i+1;
	end
	return s;
end

link(s) do
	Out.writes("<A href=\q#");
	linkify(s);
	Out.writes("\q>");
	htmlize(s);
	Out.writes("</A>");
end

name(s) do
	Out.writes("<A name=\q");
	linkify(s);
	Out.writes("\q>");
	htmlize(s);
	Out.writes("</A>");
end

setcolor(c) do var buf::100;
	Out.writes(str.format(buf, "<SPAN class=\"%s\">", [(c)]));
end

endcolor() Out.writes("</SPAN>");

literal(s) do
	setcolor(C_numlit);
	htmlize(s);
	endcolor();
end

strlit(s) do
	setcolor(C_strlit);
	htmlize(s);
	endcolor();
end

comment(s) do
	setcolor(C_comment);
	htmlize(s);
	endcolor();
end

code(s) link(s);

define(s) do
	setcolor(C_def);
	Out.writes("<B>");
	name(s);
	Out.writes("</B>");
	endcolor();
end

numeric(s, hex) do
	while (s::0) do
		if (s::0 < '0' \/ s::0 > '9') do
			if (\hex \/ s::0 < 'a' \/ s::0 > 'f')
				return 0;
		end
		s := @s::1;
	end
	return -1;
end

process() do
	var	word::MAXWORD+1;
	var	state, last, pflag, nconsts;
	var	i;

	state := 0;
	last := 0;
	pflag := 0;
	nconsts := -1;
	while (1) do
		readword(word, 0);
		if (In.eof()) leave;
		if (	\state /\
			(\str.comp(word, "\\") \/
			\str.comp(word, "\\d"))
		) do
			code(word);
			if (Rejected) Out.wrch(Rejected);
			if (Rejected \= '\n') do
				In.reads(word, MAXWORD);
				comment(word);
			end
			Rejected := 0;
			state := 0;
			last := 0;
			loop;
		end
		if (	\state /\
			(\str.comp(word, "(") \/
			\str.comp(word, "(l"))
		) do
			code(word);
			readword(word, ')');
			comment(word);
			Rejected := 0;
			literal(")");
			state := 0;
			last := 0;
			loop;
		end
		if (	\state /\
			(\str.comp(word, "abort\"") \/
			\str.comp(word, ".\"") \/
			\str.comp(word, "s\""))
		) do
			code(word);
			readword(word, '"');
			strlit(word);
			Rejected := 0;
			literal("\"");
			state := 0;
			last := 0;
			loop;
		end
		if (\state /\ \str.comp(word, "defer")) do
			code(word);
			readword(word, 0);
			Out.writes("<A name=\q");
			linkify(word);
			Out.writes("\">&nbsp;</A>");
			Out.writes("<A href=\q#(");
			linkify(word);
			Out.writes(")\q>");
			htmlize(word);
			Out.writes("</A>");
			state := 0;
			last := 0;
			loop;
		end
		if (\state /\ \str.comp(word, "alias")) do
			code(word);
			readword(word, 0);
			state := 1;
		end
		last := state;
		state :=  last-> 0:
			\str.comp(word, ":") \/
			\str.comp(word, ":h") \/
			\str.comp(word, ":o") \/
			\str.comp(word, "var") \/
			\str.comp(word, "const") \/
			\str.comp(word, "array") \/
			\str.comp(word, "user-var")
		-> 1: 0;
		if (\str.comp(word, "const"))
			nconsts := nconsts+1;
		if (\last /\ \str.comp(word, ":h")) do
			pflag := 1;
		end
		if (pflag /\ \str.comp(word, "next;")) pflag := 0;
		if (\last /\ numeric(word, pflag)) do
			ie (	\pflag /\
				(\str.comp(word, "0") /\ nconsts>=0 \/
				 \str.comp(word, "1") /\ nconsts>=1 \/
				 \str.comp(word, "2") /\ nconsts>=2)
			) do
				code(word);
			end
			else do
				literal(word);
			end
			state := 0;
			loop;
		end
		if (\last /\ (
			\str.comp(word, "bin") \/
			\str.comp(word, "char") \/
			\str.comp(word, "dec") \/
			\str.comp(word, "hex") \/
			\str.comp(word, "oct") \/
			\str.comp(word, "[bin]") \/
			\str.comp(word, "[char]") \/
			\str.comp(word, "[dec]") \/
			\str.comp(word, "[hex]") \/
			\str.comp(word, "[oct]"))
		) do
			code(word);
			readword(word, 0);
			literal(word);
			state := 0;
			loop;
		end
		ie (last)
			define(word);
		else
			code(word);
	end
end

bs(href, text) do var buf::MAXWORD;
	code("\\");
	Out.wrch('\s');
	if (href) Out.writes(str.format(buf, "<A href=\"%s\">", [(href)]));
	str.copy(buf, text);
	comment(buf);
	if (href) Out.writes("</A>");
	Out.writes("\n");
end

head() do var b::128;
	Out.writes("<!DOCTYPE HTML PUBLIC");
	Out.writes(" \"-//W3C//DTD HTML 4.0 Transitional//EN\">\n");
	Out.writes("<HTML>\n");
	Out.writes("<HEAD>\n");
	Out.writes("<TITLE>Hyperlinked SOL Source Code</TITLE>\n");
	Out.writes("<STYLE type=\"text/css\">\n");
	Out.writes("PRE.sol   { background:  #000000;\n"); 
	Out.writes("            font-size:   larger;\n");
	Out.writes("            color:       #c0c0c0;  }\n");
	Out.writes("SPAN.t    { font-style:  normal;   } /* text */\n");
	Out.writes("SPAN.n    { color:       #e0e080;  } /* literal */\n");
	Out.writes("SPAN.s    { color:       #ffa080;  } /* string */\n");
	Out.writes("SPAN.c    { font-style:  italic;    /* comment */\n");
	Out.writes("            color:       #c0c0c0;  }\n");
	Out.writes("SPAN.d    { font-weight: bold;      /* definition */\n");
	Out.writes("            color:       #a0ffa0;  }\n");
	Out.writes("A         { text-decoration: none; } /* thread */\n");
	Out.writes("A:link    { color:       #90c0ff;  } /* thread */\n");
	Out.writes("A:visited { color:       #90c0ff;  } /* thread */\n");
	Out.writes("A:hover   { color:       #000000;\n");
	Out.writes("            background:  #90c0ff;  }\n");
	Out.writes("</STYLE>\n");
	Out.writes("</HEAD>\n");
	Out.writes("<BODY>\n");
	Out.writes("<PRE class=sol>\n");
end

tail() do
	Out.writes("</SPAN></PRE>\n");
	Out.writes("</BODY>\n");
	Out.writes("</HTML>\n");
end

do
	init();
	head();
	process();
	tail();
	fini();
end

