\ Generic SOL Kernel Version 0.9.0
\ By Nils M Holm, 2001-2010 (In the Public Domain)

\ ----- bootstrap -----

    128 c,           \ HALT
     83 c,           \ magic
     79 c,
     76 c,
      0 ,            \ 'sol (entry point)
      0 ,            \ bdev (boot device)

\ ----- trap vectors -----

   1024 ,            \ divide by zero
   1024 ,            \ stack check
   1024 ,            \ rstack check
   1024 ,            \ memory check

\ ----- control -----

\d halt ( -- ) m-- | halt the machine
:h halt
   80                \ HALT
   next;

\d (:) ( ; a -- a' ) m-- | address list interpreter
:h (:)
   81                \ INTP
   next;

\d (lit) ( -- n ) m-- | push inline literal number
:h (lit)
   83                \ LIT
   next;

\d (var) ( -- a ) m-- | run time for variables
:h (var)
   9e                \ FROMR
   next;

\d (const) ( -- n ) m-- | run time for constants
:h (const)
   84                \ CONST
   next;

\d (array) ( n -- a ) m-- | run time for arrays
:h (array)
   85                \ ARRAY
   next;

\d exit ( ; a -- ) m-- | return from thread
:h exit
   86                \ EXIT
   next;

\d exec ( a -- ) m-- | execute word
:h exec
   87                \ EXEC
   next;

\d (br) ( -- ) m-- | branch to inline address
:h (br)
   88                \ BR
   next;

\d (bf) ( f -- ) m-- | branch on false to inline address
:h (bf)
   89                \ BF
   next;

\ ----- stack and memory ops -----

\d dup ( n -- n n ) m-- | duplicate tos
:h dup
   90                \ DUP
   next;

\d drop ( n -- ) m-- | discard tos
:h drop
   91                \ DROP
   next;

\d swap ( n m -- m n ) m-- | swap tos and tos-1
:h swap
   92                \ SWAP
   next;

\d over ( n m -- n m n ) m-- | duplicate tos-1
:h over
   93                \ OVER
   next;

\d rot ( n m k -- m k n ) m-- | rotate stack up
:h rot
   94                \ ROT
   next;

\d @ ( a -- n ) m-- | indirection
:h @
   95                \ GET
   next;

\d c@ ( a -- c ) m-- | byte indirection
:h c@
   96                \ CGET
   next;

\d ! ( n a -- ) m-- | store machine word
:h !
   97                \ PUT
   next;

\d c! ( c a -- ) m-- | store byte
:h c!
   98                \ CPUT
   next;

\d sp@ ( -- a ) m-- | push stack pointer
:h sp@
   99                \ SPGET
   next;

\d sp! ( a -- ? ) m-- | pop stack pointer
:h sp!
   9a                \ SPPUT
   next;

\d rp@ ( -- a ) m-- | push return stack pointer
:h rp@
   9b                \ RPGET
   next;

\d rp! ( a -- ; -- ? ) m-- | pop return stack pointer
:h rp!
   9c                \ RPPUT
   next;

\d >r ( n -- ; -- n ) m-- | transfer value to rstack
:h >r
   9d                \ TOR
   next;

\d r> ( -- n ; n -- ) m-- | transfer value from rstack
:h r>
   9e                \ FROMR
   next;

\d r@ ( -- n ) m-- | copy value from rstack
:h r@
   9f                \ RGET
   next;

\d rdrop ( -- ; n -- ) m-- | drop top of rstack
:h rdrop
   a0                \ RDROP
   next;

\ ----- arithmetics -----

\d + ( n m -- n+m ) m-- | sum
:h +
   b0                \ PLUS
   next;

\d - ( n m -- n-m ) m-- | difference
:h -
   b1                \ MINUS
   next;

\d * ( n m -- n*m ) m-- | product
:h *
   b2                \ MUL
   next;

\d u* ( u v -- u*v ) m-- | unsigned product
:h u*
   b3                \ UMUL
   next;

\d /mod ( n m -- n/m n mod m ) m-- | quotient+modulus
:h /mod
   b4                \ DIVMOD
   next;

\d u/mod ( u v -- u/v u mod v ) m-- | unsigned quot+mod
:h u/mod
   b5                \ UDIVMD
   next;

\d and ( u v -- u/\v ) m-- | bitwise logical product
:h and
   b6                \ AND
   next;

\d or ( u v -- u\/v ) m-- | bitwise logical sum
:h or
   b7                \ OR
   next;

\d xor ( u v -- u xor v ) m-- | bitwise logical exclusive or
:h xor
   b8                \ XOR
   next;

\d shl ( u v -- u shl v ) m-- | bitwise left shift
:h shl
   b9                \ SHL
   next;

\d shr ( u v -- u shr v ) m-- | bitwise right shift
:h shr
   ba                \ SHR
   next;

\d neg ( n -- -n ) m-- | negation
:h neg
   bb                \ NEG
   next;

\d inv ( u -- inv u ) m-- | bitwise complement
:h inv
   bc                \ INV
   next;

\d 1+ ( n -- n+1 ) m-- | increment
:h 1+
   bd                \ INC
   next;

\d 1- ( n -- n-1 ) m-- | decrement
:h 1-
   be                \ DEC
   next;

\d +! ( n a -- ) m-- | increment address by value
:h +!
   bf                \ ADDTO
   next;

\d bsz ( u -- v ) m-- | bitwise scan for zero
:h bsz
   c0                \ BSZ
   next;

\d not ( f -- not f ) m-- | logical complement
:h not
   c1                \ NOT
   next;

\d < ( n m -- n<m ) m-- | 'less than' predicate
:h <
   c2                \ LESS
   next;

\d > ( n m -- n>m ) m-- | 'greater than' predicate
:h >
   c3                \ GRTR
   next;

\d u< ( u v -- u<v ) m-- | 'unsigned less' predicate
:h u<
   c4                \ ULESS
   next;

\d u> ( u v -- u>v ) m-- | 'unsigned greater' predicate
:h u>
   c5                \ UGRTR
   next;

\d = ( n m -- n=m ) m-- | 'equal to' predicate
:h =
   c6                \ EQUAL
   next;

\ ----- block ops -----

\d cmove ( src dst n -- ) m-- | move memory region
:h cmove
   d0                \ CMOVE
   next;

\d cscan ( a n c -- a|0 ) m-- | search byte in memory
:h cscan
   d1                \ CSCAN
   next;

\d -text ( a1 a2 n -- n ) m-- | compare strings
:h -text
   d2                \ DTEXT
   next;

\d fill ( a n c -- ) m-- | fill memory region
:h fill
   d3                \ FILL
   next;

\d lookup ( a x -- 0|a -1 ) m-- | find lfa of lexicon entry
:h lookup
   d4                \ LOOKUP
   next;

\ ----- keyboard and video ops -----

\d (?kbrdy) ( -- f ) m-- | internal keyboard ready?
:h (?kbrdy)
   e0                \ KBRDY
   next;

\d (key) ( -- c ) m-- | read internal keyboard
:h (key)
   e1                \ KEY
   next;

\d (emit) ( c -- ) m-- | write to internal screen
:h (emit)
   e2                \ EMIT
   next;

\d cursor ( x y -- ) m-- | move internal cursor
:h cursor
   e3                \ CURSOR
   next;

\d color ( n -- ) m-- | change terminal color
:h color
   e4                \ COLOR
   next;

\d cpos ( -- x y ) m-- | return cursor position
:h cpos
   e5                \ CPOS
   next;

\d clreol ( n -- ) m-- | clear screen to end of line
:h clreol
   e6                \ CLREOL
   next;

\d clrscr ( n -- ) m-- | clear screen
:h clrscr
   e7                \ CLRSCR
   next;

\d dskio ( dev blk a op -- n ) m-- | run disk I/O request
:h dskio
   f0                \ DSKIO
   next;

\ ----- constants -----

\d 0 ( -- 0 ) c-- | constant 0
0 const 0

\d 1 ( -- 1 ) c-- | constant 1
1 const 1

\d 2 ( -- 2 ) c-- | constant 2
2 const 2

\d -1 ( -- -1 ) c-- | constant -1
1 neg const -1

\d true ( -- -1 ) c-- | logical truth
-1 const true

\d false ( -- 0 ) c-- | logical falsity
0 const false

\d bl ( -- c ) c-- | blank character
32 const bl

\d bpc ( -- n ) c-- | bytes per cell
4 const b/c

\d b/blk ( -- 1024 ) c-- | bytes per block
1024 const b/blk

\d b/r ( -- 64 ) c-- | bytes per record
64 const b/r

\d r/blk ( -- 16 ) c-- | records per block
16 const r/blk

\d b/stk ( -- 4096 ) c-- | bytes on stack
4096 const b/stk

\d memsize ( -- u ) c-- | memory size
hex 40000 const memsize

\d tib ( -- a ) c-- | terminal input buffer address
memsize 256 - const tib

\d dtb ( -- a ) c-- | disk transfer buffer address
tib b/blk - const dtb

\d fbl ( -- a ) c-- | free block list address
dtb b/blk - const fbl

\d s0 ( -- a ) a-- | bottom of data stack
fbl const s0

\d r0 ( -- a ) c-- | bottom of return stack
s0 b/stk - const r0

\d blk.fbl ( -- n ) c-- | free block list block
129 const blk.fbl

\d blk.init ( -- n ) c-- | initialization block
130 const blk.init

\d sol.kspace ( -- a ) c-- | kernel space address
hex 0400 const sol.kspace

\d 'sol ( -- a ) v-- | kernel boot address (var ptr!)
hex 0404 const 'sol

\d sol.bdev ( -- a ) v-- | boot device id (var ptr!)
hex 0408 const sol.bdev

\d sol.ksize ( -- a ) c-- | kernel size in blocks
64 const sol.ksize

\d key.up ( -- n ) c-- | key code: up arrow
256 const key.up

\d key.down ( -- n ) c-- | key code: down arrow
257 const key.down

\d key.left ( -- n ) c-- | key code: left arrow
258 const key.left

\d key.right ( -- n ) c-- | key code: right arrow
259 const key.right

\d col.user ( -- n ) v-- | color: user input
hex 03 value col.user

\d col.system ( -- n ) v-- | color: system output
hex 07 value col.system

\d col.alert ( -- n ) v-- | color: alarm messages
hex 06 value col.alert

\ ----- core words -----

\d h ( -- a ) v-- | beginning of heap
0 var h

\d here ( -- a ) t-- | beginning of heap
: here ( -- a ) h @ ;

\d allot ( n -- ) t-- | allocate bytes
: allot ( n -- ) here + h ! ;

\d , ( n -- ) t-- | compile cell
: , ( n -- ) here !  b/c allot ;

\d c, ( c -- ) t-- | compile byte
: c, ( c -- ) here c!  1 allot ;

\d top ( -- a ) t-- | top of user memoty
: top ( -- a ) rp@ b/blk - ;

\d nip ( n m -- m ) t-- | drop tos-1
: nip ( n m -- m ) swap drop ;

\d tuck ( n m -- m n m ) | tuck tos under tos-1
: tuck ( n m -- m n m ) swap over ;

\d -rot ( n m k -- k n m ) t-- | rotate stack down
: -rot ( n m k -- k n m ) rot rot ;

\d 0= ( n -- f ) t-- | test for zero
' not alias 0= ( n -- f )

\d 0< ( n -- f ) t-- | test for less than zero
: 0< ( n m -- n<m ) 0 < ;

\d 0> ( n -- f ) t-- | test for greater than zero
: 0> ( n m -- n>m ) 0 > ;

\d 2dup ( n m -- n m n m ) t-- | duplicate tos and tos-1
: 2dup ( n m -- n m n m ) over over ;

\d 2drop ( n m -- ) t-- | drop tos and tos-1
: 2drop ( n m -- ) drop drop ;

\d d= ( n m k l -- n=k and m=l ) t-- | compare two pairs
: d= ( n m k l -- n=k and m=l ) rot =  -rot =  and ;

\d cell+ ( n -- n+b/c ) t-- | advance to next cell
: cell+ ( n -- n+b/c ) b/c + ;

\d cell- ( n -- n-b/c ) t-- | go back one cell
: cell- ( n -- n-b/c ) b/c - ;

\d /cells ( u -- v ) t-- | convert bytes to cells
: /cells ( u -- v ) 3 + 2 shr ( b/c 1- + b/c / ) ;

\d cells ( u -- v ) t-- | convert cells to bytes
: cells ( u -- v ) b/c * ;

\d depth ( -- n ) t-- | push depth of stack
: depth ( -- n ) s0 cell-  sp@ -  /cells ;

\d <= ( n m -- n<=m ) t-- | 'less than/equal to' predicate
: <= ( n m -- n<=m ) > not ;

\d >= ( n m -- n>=m ) t-- | 'greater than/equal to' predicate
: >= ( n m -- n>=m ) < not ;

\d u<= ( u v -- u<=v ) t-- | 'unsigned less/equal' predicate
: u<= ( u v -- u<=v ) u> not ;

\d u>= ( u v -- u>=v ) t-- | 'unsigned greater/equal' pred.
: u>= ( u v -- u>=v ) u< not ;

\d <> ( n m -- n<>m ) t-- | 'not equal to' predicate
: <> ( u v -- u>=v ) = not ;

\d within ( n l h -- f ) t-- | range check: l <= n <= h
: within ( n l h -- f ) rot dup rot <=  -rot <=  and ;

\d uwithin ( n l h -- f ) t-- | range check: l u<= n u<= h
: uwithin ( n l h -- f ) rot dup rot u<=  -rot u<=  and ;

\d count ( a -- a n ) t-- | extract length of string
: count ( a -- a n ) dup 1+ swap c@ ;

\d s+ ( a n -- a+1 n-1 ) t-- | move to next char of string
: s+ ( a n -- a+1 n-1 ) 1- swap 1+ swap ;

\d 2+ ( n -- n+2 ) t-- | increment by 2
: 2+ ( n -- n+2 ) 1+ 1+ ;

\d 2- ( n -- n-2 ) t-- | decrement by 2
: 2- ( n -- n-2 ) 1- 1- ;

\d / ( n m -- n/m ) t-- | signed quotient
: / ( n m -- n/m ) /mod drop ;

\d mod ( n m -- n mod m ) t-- | modulus
: mod ( n m -- n mod m ) /mod nip ;

\d u/ ( u -- u/v ) t-- | unsigned quotient
: u/ ( u v -- u/v ) u/mod drop ;

\d umod ( u v -- u mod v ) t-- | unsigned modulus
: umod ( u v -- u mod v ) u/mod nip ;

\d abs ( n -- |n| ) t-- | absolute value
: abs ( n -- |n| ) dup 0< if  neg then ;

\d min ( n m -- n|m ) t-- | minimum
: min ( n m -- n|m ) 2dup < if  drop else  nip then ;

\d max ( n m -- n|m ) t-- | maximum
: max ( n m -- n|m ) 2dup < if  nip else  drop then ;

\ ----- lexicon functions -----

\d lexicon ( -- a ) v-- | pointer to lexicon
0 var lexicon

\d last ( -- a ) v-- | ptr to last created word
0 var last

\d fence ( -- a ) v-- | top of protected area
0 var fence

\d >dict ( -- a ) v-- | beginning of dictionary
0 var >dict

\d l>code ( a -- a ) t-- | move from link to code field
: l>code ( a -- a ) cell+ dup c@ 127 and + 2+ ;

\d p>code ( a -- a ) t-- | move from parameter to code field
' cell- alias p>code ( a -- a )

\d >name ( a -- a ) t-- | move from code to name field
: >name ( a -- a ) 1- dup c@ - 1- ;

\d >link ( a -- a ) t-- | move from code to link field
: >link ( a -- a ) >name cell- ;

\d >pfa ( a -- a ) t-- | move from code to parameter field
: >pfa ( a -- a ) cell+ ;

\ ----- compilation -----

\d lit ( n -- ) t-- | compile literal
: lit ( n -- ) (lit) (lit) , , ;

\d syntax ( -- ) t-- | make last word syntax word
: syntax ( -- ) last @ l>code >name  dup c@ 128 or  swap c! ;

defer '

\d escape ( -- ) tsi | compile syntax word
: escape ( -- ) ' , ; syntax

\d ['] ( -- a ) tsi | compile code address of word
: ['] ( -- a ) ' lit ; syntax

\d compile ( -- ) tsi | compile code to compile word
: compile ( -- ) escape [']  ['] , , ; syntax

\ ----- flow control -----

\d recurse ( -- ) ts- | call current definition
: recurse ( -- ) last @ l>code , ; syntax

\d mark ( -- a ) t-- | generate mark for backpatching
: mark ( -- a ) here  0 , ;

\d resolve ( a -- ) t-- | backpatch marked address
: resolve ( a -- ) here swap ! ;

\d if ( f -- ) ts- | introduce if-else-then
: if ( f -- ) compile (bf) mark ; syntax

\d else ( -- ) ts- | else part of if-else-then
: else ( -- ) compile (br) mark  swap resolve ; syntax

\d then ( -- ) ts- | delimit if-else-then
: then ( -- ) resolve ; syntax

\d begin ( -- ) ts- | introduce loop
: begin ( -- ) here ; syntax

\d while ( f -- ) ts- | condition of begin-while-repeat
: while ( f -- ) escape if ; syntax

\d repeat ( -- ) ts- | delimit begin-while-repeat
: repeat ( -- ) compile (br) swap ,  resolve ; syntax

\d again ( -- ) ts- | delimit begin-again
: again ( -- ) compile (br) , ; syntax

\d until ( f -- ) ts- | delimit begin-until
: until ( f -- ) compile (bf) , ; syntax

\d for ( n -- ) ts- | introduce for-next
: for ( n -- ) compile >r  here ; syntax

\d (next) ( -- f ; n -- [n] ) t-- | runtime of next
: (next) ( -- f ; n -- [n] ) r>  r> 1- dup if
    >r false else  drop true then  swap >r ;

\d next ( -- ) ts- | delimit for-next
: next ( -- ) compile (next)  compile (bf) , ; syntax

\d (do) ( i' i -- ; -- i' i ) t-- | runtime of do
: (do) ( i' i -- ; -- i' i ) swap  r> -rot  >r >r >r ;

\d do ( i' i -- ) ts- | introduce do-loop
: do ( i' i -- ) compile (do)  here ; syntax

\d (loop) ( -- f ; i' i -- [i' i] ) t-- | runtime of loop
: (loop) ( -- f ; i' i -- [i' i] ) r>  r> 1+ r> 2dup u> if
    2drop true else  >r >r false then  swap >r ;

\d loop ( -- ) ts- | delimit do-loop
: loop ( -- ) compile (loop)  compile (bf) , ; syntax

\d i ( -- i ) t-- | for-next/do-loop index
: i ( -- i ) r> r@ swap >r ;

\d i' ( -- i' ) t-- | do-loop limit
: i' ( -- i' ) r> r> r@ -rot >r >r ;

\d break ( ; n -- 1 ) t-- | leave for-next
: break ( ; n -- 1 ) r> rdrop 1 >r >r ;

\d leave ( ; i' i -- i' i' ) t-- | leave do-loop
: leave ( -- i' i -- i' i' ) r> r> r> nip dup >r >r >r ;

\d unfor ( ; i -- ) t-- | remove for-next parameter
' rdrop alias unfor ( ; i -- )

\d unloop ( ; i' i -- ) t-- | remove do-loop parameters
: unloop ( -- ) r> rdrop rdrop >r ;

\d case ( -- ) ts- | introduce case
: case ( -- ) 0 ; syntax

\d => ( f -- ) ts- | introduce case of case
: => ( f -- ) escape if ; syntax

\d ;; ( -- ) ts- | delimit case of case
: ;; ( -- ) escape else ; syntax

\d else> ( -- ) ts- | introduce default of case
: else> ( -- ) ; syntax

\d end ( -- ) ts- | delimit case
: end ( -- ) begin  dup while
    resolve repeat  drop ; syntax

\d ?dup ( n -- n [n] ) t-- | dup if tos not zero
: ?dup ( -- ) dup if  dup then ;

\d ?if ( f -- f| ) ts- | short form of ?dup if
: ?if ( f -- f| ) compile ?dup  escape if ; syntax

\d ?while ( f -- f| ) ts- | short form of ?dup while
: ?while ( f -- f| ) compile ?dup  escape while ; syntax

\d ?break ( f -- ) t-- | break if tos true
: ?break ( f -- ) if  r> break >r then ;

\d ?leave ( f -- ) t-- | leave if tos true
: ?leave ( f -- ) if  r> leave >r then ;

\d ?exit ( f -- ) t-- | exit if tos true
: ?exit ( f -- ) if  rdrop then ;

\ ----- console i/o -----

\d 'key ( -- a ) v-- | key hook
' (key) var 'key

\d kbmap ( c -- c' ) t-- | map raw key to key code
: kbmap ( c -- c' )
   dup 328 = if drop key.up    then
   dup 336 = if drop key.down  then
   dup 333 = if drop key.right then
   dup 331 = if drop key.left  then ;

\d key ( -- c ) t-- | receive character from user tty
: key ( -- c ) 'key @ exec kbmap ;

\d '?key ( -- a ) v-- | ?key hook
' (?kbrdy) var '?key

\d ?key ( -- f ) t-- | key pending?
: ?key ( -- f ) '?key @ exec ;

\d 'emit ( -- a ) v-- | emit hook
' (emit) var 'emit

\d emit ( c -- ) t-- | transmit character to user tty
: emit ( c -- ) 'emit @ exec ;

\d usrtty ( -- a ) a-- | user tty routines
0 array usrtty  ' (key) ,  ' (?kbrdy) ,  ' (emit) ,

\d space ( -- ) t-- | emit space character
: space ( -- ) bl emit ;

\d spaces ( n -- ) t-- | emit space characters
: spaces ( n -- ) for space next ;

\d tty! ( -- ) t-- | establish default tty hooks
: tty! ( -- ) 0 usrtty @ 'key !  1 usrtty @ '?key !
  2 usrtty @ 'emit ! ;

\d bs ( -- ) t-- | emit backspace character
: bs ( -- ) 8 emit ;

\d rubout ( -- ) t-- | emit destructive backspace
: rubout ( -- ) bs space bs ;

\d cr ( -- ) t-- | emit newline sequence
: cr ( -- ) 10 emit 13 emit ;

\d home ( -- ) t-- | move cursor to top/left corner
: home ( -- ) 0 0 cursor ;

\d page ( -- ) t-- | clear screen
: page ( -- ) home clrscr ;

\d so ( -- ) t-- | turn on standout mode
: so ( -- ) [hex] 70 color ;

\d se ( -- ) t-- | turn off standout mode
: se ( -- ) [hex] 07 color ;

\d type ( a n -- ) t-- | emit string
: type ( a n -- ) begin  dup while
    over c@ emit  s+  repeat 2drop ;

defer quit

\d what? ( a -- ) t-- | emit string and trailing ? and cr
: what? ( a -- ) col.alert color  space  count type
  [char] ? emit  cr  quit ;

\ ----- scanner -----

\d >buf ( -- a ) v-- | pointer to input buffer
0 var >buf

\d >in ( -- a ) v-- | input pointer
0 var >in

\d >lim ( -- a ) v-- | pointer to input limit
0 var >lim

\d len ( -- a ) v-- | length of text in tib
0 var len

\d reread ( -- ) t-i | re-read input buffer
: reread ( -- ) >buf @ >in ! ;

\d reject ( -- ) t-i | un-read recently read character
: reject ( -- ) -1 >in +! ;

\d >in@ ( -- c ) t-i | read character from input buffer
: >in@ ( -- c ) >in @  >lim @  u>= if  0 exit then
  >in @ c@  1 >in +! ;

\d skip ( c -- ) t-i | skip characters in input buffer
: skip ( c -- ) begin  >in@ over <> until  drop  reject ;

\d word ( c -- a ) t-i | extract word from input buffer
: word ( c -- a ) here >r  0 , 0 c,  dup skip
  begin  >in@ 2dup <>  over and  while  c, repeat
  2drop  here r@ - cell- 1-  r@ cell+ c!  r> dup h !  cell+
  ;

\ ----- character functions -----

\d char ( -- c ) t-i | push character literal
: char ( -- c ) bl word 1+ c@ ;

\d [char] ( -- c ) tsi | compile character literal
: [char] ( -- c ) char lit ; syntax

\d >upper ( c -- c2 ) t-- | compile character literal
: >upper ( c -- c2 ) dup [char] a [char] z within if
    32 - then ;

\d >lower ( c -- c2 ) t-- | convert character to lower case
: >lower ( c -- c2 ) dup [char] A [char] Z within if
    32 + then ;

\d >visual ( c -- c2 ) t-- | replace unprintable characters
: >visual ( c -- c2 ) dup  bl [char] ~ within not  if
   drop [char] . then ;

\ ----- string/memory functions -----

\d (skip) ( a -- a r ) t-- | skip over inline string literal
: (skip) ( a -- a r ) dup  dup c@ + 1+ ;

\d (s") ( -- a ) t-- | runtime of s"
: (s") ( -- a ) r> (skip) >r ;

\d (save") ( -- a ) t-i | save string in lexicon
: (save") ( -- a ) here  >in@ [char] " = if
    here  0 , 0 c,  h !  0  else
    reject  [char] " word c@ then  cell+ 1+ allot ;

\d s" ( -- a ) tsi | compile string literal
: s" ( -- a ) (save")  ['] (s") swap ! ; syntax

\d (.") ( -- ) t-- | runtime of ."
: (.") ( -- ) r> (skip) >r count type ;

\d ." ( -- ) tsi | type inline string literal
: ." ( -- ) (save")  ['] (.") swap ! ; syntax

\d (abort") ( a -- ) t-- | runtime of abort"
: (abort") ( a -- ) tty!  r> (skip) >r  what? ;

\d abort" ( -- ) tsi | print message and quit
: abort" ( -- ) (save")  ['] (abort") swap ! ; syntax

\d upcase ( a -- ) t-- | convert string to upper case
: upcase ( a -- ) count begin  dup while
    over dup c@  >upper swap c!  s+  repeat 2drop ;

\d locase ( a -- ) t-- | convert string to lower case
: locase ( a -- ) count begin dup while
    over dup c@  >lower  swap c!  s+  repeat  2drop ;

\d blank ( a n -- ) t-- | fill memory with blanks
: blank ( a n -- ) bl fill ;

\d erase ( a n -- ) t-- | fill memory with zeroes
: erase ( a n -- ) 0 fill ;

\d .( ( -- ) tsi | print execution time comment
: .( ( -- ) [char] ) word count type ; syntax

\d ( ( -- ) tsi | introduce parenthesized comment
: ( ( -- ) [char] ) word drop ; syntax

\d \ ( -- ) tsi | introduce comment to end of line
: \ ( -- ) >in @  b/r +  b/r 1- inv and  >in ! ; syntax

\d \d ( -- ) t-i | introduce documentation comment
: \d ( -- ) escape \ ;

\d \s ( -- ) t-i | introduce comment to end of screen
: \s ( -- ) 0 word drop ;

\ ----- more lexicon functions -----

\d create ( -- ) t-i | create lexicon entry
: create ( -- ) bl word  dup locase ( w)
  lexicon @ here !  here last !
  here l>code 1- h !  ( w) c@ c,  ['] (var) , ;

\d link ( -- ) t-- | link new entry into lexicon
: link ( -- ) last @ lexicon ! ;

\d find ( a -- 0|a -1 ) t-i | find code addr of lexicon entry
: find ( a -- 0|a -1 ) dup locase  lexicon @ lookup dup if
    swap l>code swap then ;

\d ' ( -- a ) t-i | find code address of lexicon entry or fail
: (') ( -- a ) bl word  dup find if  nip else  what? then ;
' (') is '

\d forget ( -- ) t-i | delete words from lexicon
: forget ( -- ) ' >link  dup fence @ u<= if
   abort" fence" then  dup @ lexicon !  dup last !  h ! ;

\d empty ( -- ) t-- | forget all unprotected words
: empty ( -- ) here last @  begin  dup while
    dup @ fence @ u< if  dup lexicon !  last !  h !  exit then
    nip dup @ repeat  2drop ;

\d dict> ( -- ) t-- | begin new dictionary
: dict> ( -- ) last @ >dict ! ;

\d <dict ( -- ) t-i | delimit and define dictionary
: <dict ( -- ) create  >dict @ ,  lexicon @ @ ,  link ;

\d show ( a -- ) t-- | show (activate) dictionary
: show ( a -- ) dup cell+ @  swap p>code >link @ ! ;

\d hide ( a -- ) t-- | hide (deactivate) dictionary
: hide ( a -- ) dup @  swap p>code >link @ ! ;

\ ----- definition words -----

\d var ( n -- ) t-i | define variable
: var ( n -- ) create , link ;

\d patch ( a -- ) t-- | patch code field
: patch ( a -- ) here cell- ! ;

\d const ( n -- ) t-i | define constant
: const ( n -- ) create  ['] (const) patch  ,  link ;

\d value ( n -- ) t-i | define user variable
' const alias value ( n -- )

\d array ( n -- ) t-i | define array
: array ( n -- ) create  ['] (array) patch
  cells allot  link ;

\d alias ( a -- ) t-i | define alias
: alias ( a -- ) create  here cell- h !
  [hex] 8a ( REDIR ) c, , link ;

\ ----- numeric i/o and converion -----

\d base ( -- a ) v-- | input/output radix
10 var base

\d hld ( -- a ) v-- | ptr to numeric conversion buffer
0 var hld

\d ?# ( c -- 0|n -1 ) t-- | test if character is a digit
: ?# ( c -- 0|n -1 ) dup [char] 0 [char] 9 within if
    [char] 0 - true exit then
  base @ 10 > if  dup [char] a [char] f within if
    [char] W - true exit then  then  drop false ;

\d ?num ( a -- 0|n -1 ) t-- | test if string is numeric
: ?num ( a -- 0|n -1 ) 0 >r  count begin  dup while
    over c@ ?# not if  2drop  rdrop  false exit then
    r> base @ * + >r  s+
  repeat  2drop  r> true ;

\d pad ( -- a ) a-- | scratch pad address
: pad ( -- a ) here 128 + ;

\d <# ( n m -- n m ) t-- | begin pictured conversion
: <# ( n m -- n m ) pad 128 +  hld ! ;

\d hold ( c -- ) t-- | insert character into scratch pad
: hold ( c -- ) hld @ c!  -1 hld +! ;

\d sign ( n m -- n m ) t-- | hold minus if n negative
: sign ( n m -- n m ) over 0< if  [char] - hold then ;

\d # ( n -- m ) t-- | convert and hold digit
: # ( n -- m ) base @ u/mod  [char] 0 +
  dup [char] 9 > if  39 + then  hold ;

\d #s ( n -- 0 ) t-- | convert and hold remaining digits
: #s ( n -- 0 ) begin  # dup 0= until ;

\d #> ( n n -- a n ) t-- | end pictured conversion
: #> ( n n -- a n ) 2drop  hld @ 1+  pad 128 + hld @ - ;

\d . ( n -- ) t-- | print signed number
: . ( n -- ) space  dup abs <# #s sign #> type ;

\d u. ( u -- ) t-- | print unsigned number
: u. ( u -- ) space  dup <# #s #> type ;

\d .pad ( l a n -- a n ) t-- | print l-n spaces if l-n>0
: .pad ( l a n -- a n ) rot over -  dup 0> if
    spaces else  drop then ;

\d .r ( n len -- ) t-- | print signed number w/padding
: .r ( n len -- ) swap  dup abs <# #s sign #> .pad type ;

\d u.r ( u len -- ) t-- | print unsigned number w/padding
: u.r ( u len -- ) swap  dup <# #s #> .pad type ;

\d ? ( a -- ) t-- | print value of cell
: ? ( a -- ) @ . ;

\d u? ( a -- ) t-- | print unsigned value of cell
: u? ( a -- ) @ u. ;

\d c? ( a -- ) t-- | print value of character cell
: c? ( a -- ) c@ . ;

\d bin! ( -- ) t-- | set numeric base to 2
: bin! ( -- ) 2 base ! ;

\d dec! ( -- ) t-- | set numeric base to 10
: dec! ( -- ) 10 base ! ;

\d hex! ( -- ) t-- | set numeric base to 16
: hex! ( -- ) 16 base ! ;

\d #in ( -- n ) t-i | scan and translate number
: #in ( -- n ) bl word  ?num not if  0 then ;

\d nbase ( n -- n ) t-i | push base-n literal
: nbase ( n -- n ) base @ swap  base !
  bl word  ?num not if
     base !  abort" bad number" then  swap base ! ;

\d (nbase) ( n -- ) t-i | runtime of '[bin]' and friends
: (nbase) ( n -- ) nbase lit ;

\d [bin] ( -- n ) tsi | compile binary literal
: [bin] ( -- n ) 2 (nbase) ; syntax

\d [dec] ( -- n ) tsi | compile decimal literal
: [dec] ( -- n ) 10 (nbase) ; syntax

\d [hex] ( -- n ) tsi | compile hexa-decimal literal
: [hex] ( -- n ) 16 (nbase) ; syntax

\d bin ( -- n ) t-i | push binary number
: bin ( -- n ) 2 nbase ;

\d dec ( -- n ) t-i | push decimal number
: dec ( -- n ) 10 nbase ;

\d hex ( -- n ) t-i | push hexa-decimal number
: hex ( -- n ) 16 nbase ;

\ ----- low level block i/o -----

\d dsk.dev ( -- a ) v-- | disk device id
0 var dsk.dev

\d dsk.rdop ( -- n ) c-- | disk read operator
0 const dsk.rdop

\d dsk.wrop ( -- n ) c-- | disk write operator
1 const dsk.wrop

\d i/o? ( f -- ) t-- | test for i/o error
: i/o? ( f -- ) if  abort" i/o error" then ;

\d dsk-i/o ( dev blk a op -- n ) t-- | disk i/o request handler
: dsk-i/o ( dev blk a op -- n ) >r dsk.dev @ -rot r>
  dskio i/o? ;

\d 'read ( -- a ) v-- | block read hook
0 var 'read

\d 'write ( -- a ) v-- | block write hook
0 var 'write

\d dskrd ( blk a -- ) t-- | read disk block
: dskrd ( blk a -- ) dsk.rdop  dsk-i/o ;

\d dskwr ( blk a -- ) t-- | write disk block
: dskwr ( blk a -- ) dsk.wrop dsk-i/o ;

\d (dsk!) ( dev -- ) t-- | select disk device w/o sync
: (dsk!) ( dev -- ) dsk.dev !  blk.fbl fbl dskrd
  ['] dskrd 'read !  ['] dskwr 'write ! ;

\d dsk! ( dev -- ) t-- | select disk device w/ sync
: dsk! ( dev -- ) blk.fbl fbl dskwr  (dsk!) ;

\ ----- abstract block i/o -----

\d blk ( -- a ) v-- | block currently in dtb
-1 var blk

\d dirty ( -- a ) v-- | block in dtb changed?
false var dirty

\d safe ( -- a ) v-- | safe mode (auto-sync free list)
false var safe

\d read ( blk a -- ) t-- | read block
: read ( blk a -- ) 'read @ exec ;

\d write ( blk a -- ) t-- | write block
: write ( blk a -- ) 'write @ exec ;

\d flush ( -- ) t-- | re-write block in dtb, if dirty
: flush ( -- ) dirty @ if  blk @ dtb write then
  false dirty ! ;

\d update ( -- ) t-- | mark block in dtb dirty
: update ( -- ) true dirty ! ;

\d discard ( -- ) t-- | clear block buffer
: discard ( -- ) -1 blk !  false dirty ! ;

\d block ( blk -- a ) t-- | read block into dtb
: block ( blk -- a ) flush  dup blk @ = if  drop else
    dup dtb read  blk ! then  dtb ;

\d save-kernel ( -- ) t-- | save kernel
: save-kernel ( -- ) sol.kspace
  sol.ksize 1 do  i dup . over write  b/blk +  loop  drop ;

\d src@ ( -- src[4] ) t-- | save current input parameters
: src@ ( -- src[4] ) blk @  >buf @  >lim @  >in @ ;

\d src! ( src[4] -- ) t-- | restore input parameters
: src! ( src[4] -- ) >in !  >lim !  >buf !
  dup -1 = if  blk ! else  block drop then ;

\d sync ( -- ) t-- | update free list
: sync ( -- ) blk.fbl fbl write ;

\d auto-sync ( -- ) t-- | automatically update free list
: auto-sync ( -- ) safe @ if  sync then ;

\d fbaddr ( blk -- a m ) t-- | get freelist bit address
: fbaddr ( blk -- a m ) 8 /mod  1 swap shl  swap fbl + swap ;

\d freebit ( a m -- a m f ) t-- | extract bit from free list
: freebit ( a m -- a m f ) over c@ over and ;

\d ?used ( blk -- f ) t-- | test block allocation
: ?used ( blk -- f ) fbaddr swap c@ and 0= not ;

\d dsize ( -- n ) t-- | compute #blocks on disk
: dsize ( -- n ) 8192  ( XXX hack me harder ) ;

\d nfree ( size -- n ) t-- | retrieve number of free blocks
: nfree ( size -- n ) dsize  dup for
    i 1- ?used if  1- then  next ;

\d allocate ( blk -- ) t-- | allocate disk block
: allocate ( blk -- ) fbaddr
  freebit if  abort" used block" then
  over c@ or  swap c!  auto-sync ;

\d release ( blk -- ) t-- | release block to free list
: release ( blk -- ) fbaddr
  freebit 0= if  abort" free block" then
  inv over c@ and  swap c!  auto-sync ;

\d newblk ( -- blk ) t-- | find and allocate free block
: newblk ( -- blk ) fbl b/blk + 1-  fbl  do
    i c@ [hex] ff <> if
      i fbl - 8 *  i c@ bsz +  dup allocate
      auto-sync  unloop  exit  then
  loop  abort" disk full" ;

\d index ( lo hi -- ) t-- | list block headings
: index ( lo hi -- ) swap do  i block drop
  dtb c@  [char] ( = if
    cr  i 5 .r  space  dtb b/r type  then  loop ;

\ ----- compilers -----

\d ; ( -- ) ts- | delimit colon definition
: ; ( -- ) compile exit  link  rdrop ; syntax

\d :: ( -- ) t-i | create colon header
: :: ( -- ) create  ['] (:) patch ;

\d ] ( -- ) t-i | thread compiler
: ] ( -- ) begin  bl word  dup c@ 0= if  drop exit then
    dup find if  dup >name c@ 128 and if
        nip exec else  , drop then  else
      dup ?num if  lit drop else  what? then  then  again ;

\d [ ( -- ) ts- | exit from thread compiler
: [ ( -- ) rdrop ; syntax

\d : ( -- ) t-i | introduce colon definition
: : ( -- ) ::  ] ;

\d (defer) ( -- ) t-- | deferred word handler
: (defer) ( -- ) abort" deferred" ;

\d defer ( -- ) t-i | create deferred definition
: defer ( -- ) ::  compile (defer)  escape ; ;

\d is ( a -- ) t-i | backpatch word / change value
: is ( a -- ) ' >pfa ! ;

\d h] ( n -- ) t-i | base-16 machine code compiler
: h] ( -- ) hex!  begin bl word
    dup c@ 0= if  drop dec! exit then
    dup ?num if  c, drop else  dup find if
      nip exec else  what? then  then  again ;

\d :h ( -- ) t-i | compile hex code
: :h ( -- ) create  here p>code h !  h] ;

\d next, ( -- ) t-- | compile 'next' primitive
: next, ( -- ) [hex] 82 ( NEXT ) c, ;

\d next; ( n -- ) t-- | compile 'next' and exit :h
: next; ( n -- ) next,  dec!  link  rdrop ;

\ :run ( -- ) t-i | run compile-time word in the interpreter
: :run ( -- )  h @ >r  compile (:)  ]  compile exit
  r@ exec  r> h ! ;

\ ----- line editor -----

\d q.back ( -- ) t-- | query: move backward
: q.back ( -- ) >in @  >buf @ > if
    -1 >in +!  bs  then ;

\d q.fwd ( -- ) t-- | query: move forward
: q.fwd ( -- ) >in @  >buf @ len @ + < if
    >in @ c@ emit  1 >in +!  then ;

\d q.len ( -- n ) t-- | query: length of rest of line
: q.len ( -- n ) len @ >in @ - >buf @ + ;

\d q.type ( f -- ) t-- | query: type rest of line
: q.type ( f -- ) >in @  q.len type  space bs  ( f ) if
    q.len  ?dup if  for bs next then  then ;

\d q.ins ( c -- ) t-- | query: insert and emit character
: q.ins ( c -- ) >buf @ len @ +  >lim @ u>=  if
    drop exit then
  >in @  dup 1+  q.len ?dup if  cmove else  2drop then
  dup >in @ c!  1 >in +!  1 len +!  emit ;

\d q.del ( -- ) t-- | query: delete character left
: q.del ( -- ) >in @ >buf @ u<= ?exit  len @ 1 < ?exit
  >in @  dup 1-  q.len cmove  bs  -1 >in +!  -1 len +! ;

\d q.eol ( -- ) t-- | query: goto eol
: q.eol ( -- ) 0 q.type  >buf @ len @ + >in ! ;

\d q.edit ( n -- ) t-- | interactive line editor
: q.edit ( n -- ) len !  q.eol  begin  key  case
    dup key.left =  => drop  q.back ;;
    dup key.right = => drop  q.fwd ;;
    dup 1 =         => drop  begin >in @ >buf @  > while
                       q.back repeat ;;
    dup 5 =         => drop  q.eol ;;
    dup 8 =         => drop  q.del 1 q.type ;;
    dup 13 =        => drop  q.eol  >buf @ len @ +  >in !
                       0  >in @  c!  1 >in +!  exit ;;
    dup 21 =        => drop  q.eol ."  ^U" cr
                       >buf @ >in !  0 len ! ;;
                 else> q.ins 1 q.type
  end  again ;

\d query ( -- ) t-- | read string into tib
: query ( -- ) tib >buf !  tib >in !  tib b/r + >lim !
  0 q.edit ;

\d equery ( -- ) t-- | edit string
: equery ( -- ) >buf @  dup dup b/r + 1-  swap do
    i c@  bl <> if  drop i then  loop  >buf @ -
  dup if 1+ then  q.edit ;

\ ----- interpreter -----

\d .ok ( -- ) t-- | emit ok prompt
: .ok ( -- ) col.system color  ."  ok" cr ;

\d ?stack ( -- ) t-- | check for stack limit violations
: ?stack ( -- )
  sp@  s0 4096 -  s0 uwithin
  rp@  r0 4096 -  r0 uwithin  and not if
    abort" stack" then ;

\d interpret ( -- ) t-i | interpreter
: interpret ( -- ) col.system color  begin  bl word dup
  c@ while  dup find if  nip  exec else
      dup ?num 0= if what? then  nip then
    ?stack  repeat  drop ;

\d load ( n -- ) t-- | load and interpret block
: load ( n -- ) >r src@  r> block
  dup >buf !  dup >in !  b/blk + >lim !
  interpret  src! ;

\ ----- top level loop -----

\d quit ( -- ) t-i | top level loop
: (quit) ( -- ) r0 rp!  s0 sp!  begin
    tty!  col.user color  query  >in @ >lim !  reread
    interpret .ok  again ;
' (quit) is quit

\ ----- tools -----

\d .s ( -- ) t-- | print stack
: .s ( -- ) s0 cell- begin  dup sp@ cell+ u> while
   dup @ . cell- repeat  drop ;

\d parse ( a n c -- a ) t-i | extract word, c = delimiter
: parse ( a n c -- a ) >in @ >r  >lim @ >r
  -rot over + >lim !  >in !  word
  r> >lim !  r> >in ! ;

\d .w ( a -- ) t-- | print name of word
: .w ( a -- ) >name count 127 and type ;

\d words ( -- ) t-- | print n-words chunks of lexicon
: words ( -- ) 100  lexicon @ begin  dup while
    dup l>code space .w  swap 1- swap over  0= if
      key 13 <> if  2drop exit then  swap drop 100 swap
    then  @ repeat  2drop ;

\d hdump ( a -- ) t-- | print 16-byte hex dump
: hdump ( a -- ) base @ swap  hex!  15 0 do
    dup i + c@ 3 u.r  i 7 = if  space then
  loop  drop  base ! ;

\d cdump ( a -- ) t-- | print 16-character ascii dump
: cdump ( a -- ) 15 0 do
    dup i + c@ >visual emit loop  drop ;

\d dump ( a n -- ) t-- | print hex/char dump
: dump ( a n -- ) base @ -rot  hex!   16 /  0 do
    cr  dup  i 16 * +  dup 5 u.r  space
    dup hdump  2 spaces  cdump  loop  drop  base ! ;

\ ----- block editor -----

\d .l ( a n -- ) t-- | print n'th line of buffer at a
: .l ( a n -- ) dup 1+ 2 .r space
  b/r * + b/r type [char] | emit ;

\d list ( blk -- ) t-- | list content of block
: list ( blk -- ) block drop  15 0 do
    cr  dtb i .l  loop ;

\d p ( -- ) t-- | list most recently accessed block
: p ( -- ) blk @ list ;

\d rest ( a n -- a' n' ) t-- | compute rest of block
: rest ( a n -- a' n' ) dup  b/r *  b/blk swap -  -rot
  b/r * +  swap ;

\d delete ( a n -- ) t-- | delete line of block
: delete ( a n -- ) over >r  rest
  over b/r +  -rot  b/r - cmove
  r> r/blk 1- b/r * +  b/r  blank ;

\d insert ( a n -- ) t-- | insert blank line into block
: insert ( a n -- ) rest  dup b/r <> if over >r
    over b/r +  swap  b/r - cmove  r> else  drop then
  b/r blank ;

\d del ( n -- ) t-- | delete line n of current block
: del ( n -- ) dtb swap 1- delete ;

\d ins ( n -- ) t-- | insert blank line n into current block
: ins ( n -- ) dtb swap 1- insert ;

\d copy ( a n -- ) t-- | copy line n from current block
: copy ( a n -- ) 1- b/r * dtb +  swap  b/r cmove ;

\d subst ( a n -- ) t-- | substitute line n of current block
: subst ( a n -- ) 1- b/r * dtb +  b/r cmove ;

\d 0pad ( a -- ) t-- | pad 0-terminated record with blanks
: 0pad ( a -- ) dup >r b/r  0 cscan ?dup if
      dup r@ - b/r swap - blank then  rdrop ;

\d repl ( n -- ) t-- | replace line in current block
: repl ( n -- ) dup >r 1-  cr  dtb swap .l  cr  3 spaces
  >in @  >lim @  >buf @
  pad >in !  pad >buf !  pad b/r + >lim !
  pad r@ copy  equery  pad c@ if
    pad 0pad  pad r@ subst then
  >buf !  >lim !  >in !  rdrop ;

\ ----- linked block support -----

\d <l ( -- a ) v-- | previous block in l-block list
0 var <l

\d l> ( -- a ) v-- | next block in l-block list
0 var l>

\d (l ( -- ) t-i | l-block header
: (l ( -- ) base @ dec!  dtb 2+ >in !  bl word ?num if
    <l !  bl word ?num if
      l> !  [char] ) word drop  base !  exit  then
  then  base !  abort"  bad (l header" ;

\d l@ ( n -- 0|n -1 ) t-- | extract field from (l header
: l@ ( n -- 0|n -1 ) dtb +  10 bl parse  ?num ;

\d lhdr ( -- f ) t-- | test and read l-block header
: lhdr ( -- f ) 0 <l !  0 l> !
  dtb c@  [char] ( <>  dtb 1+ c@ [char] l <>  or if  0 else
    3 l@ if  <l ! then  9 l@ if  l> ! then  1 then ;

\d ?lhdr ( -- ) t-- | force l-block header
: ?lhdr ( -- ) lhdr ?exit  abort"  no (l header" ;

\d l! ( blk off -- ) t-- | link block at offset in l-hdr
: l! ( blk off -- ) dtb + >r  dup <# # # # # # #>
  r> swap  cmove ;

\d 2l! ( prev next -- ) t-- | update l-block links
: 2l! ( prev next -- ) 9 l!  3 l! ;

\d lindex ( lo hi -- ) t-- | list l-block headings
: lindex ( lo hi -- ) swap do  i block drop  lhdr if
      <l @ 0= if  cr  i 5 .r  space  dtb b/r type then
    then  loop ;

\d lload ( blk -- ) t-- | load l-blocks
: lload ( blk -- ) >r src@ r>  begin  block
    ?lhdr  dup >buf !  dup >in !  b/blk + >lim !
    interpret  l> @  dup 0= until  drop  src! ;

\d lclear ( blk -- ) t-- | create initial l-block
: lclear ( blk -- ) block b/blk blank
  s" (l 00000 00000  )" count  dtb swap  cmove ;

\d lblock ( prev next blk -- ) t-- | insert empty l-block
: lblock ( prev next blk -- ) dup lclear  -rot
  2dup 2l!  update  ?dup if
    block drop  ?lhdr  over l> @  2l!  update  then
  ?dup if
    block drop  ?lhdr  <l @ over  2l!  update  then
  block drop ;

\d p<l ( -- a ) v-- | old <l, used to unlink l-blocks
0 var p<l

\d pl> ( -- a ) v-- | old l>, used to unlink l-blocks
0 var pl>

\d lfree ( blk -- ) t-- | unlink l-block
: lfree ( blk -- ) block drop  ?lhdr  dtb b/r blank  update
  <l @ p<l !  l> @ pl> !  p<l @ ?dup if
    block drop  ?lhdr  <l @ pl> @ 2l!  update then
  pl> @ ?dup if
    block drop  ?lhdr  p<l @ l> @ 2l!  update then ;

\ ----- trap handler -----

\d trap ( -- ) t-- | default trap handler
: trap ( -- ) abort" trap" ;

\d trap.div ( -- a ) c-- | address of divide trap vector
hex 040c const trap.div

\d trapinit ( -- ) t-- | setup default trap handlers
: trapinit ( -- )  ['] trap trap.div ! ;

\ ----- startup and shutdown routines -----

\d bye ( -- ) t-- | shutdown routine
: bye ( -- ) flush  sync  ."  bye!" cr  halt ;

\d sol ( -- ) t-- | startup routine
: sol ( -- )
  trapinit
  cr ." SOL-3 version 0.9.0 by Nils M Holm, 2010" cr cr
  case sol.bdev @ 0= => 0 (dsk!) ;;
                  else> ." unknown boot device" cr  halt  end
  blk.init load  quit ;

\ ----- setup -----

' sol 'sol !
' sol >link lexicon !
' sol >link last !
' sol >link fence !
here h !
